
Object subclass: #ExternalDataManager
  instanceVariableNames: 
    ' dictionary stream last workStream '
  classVariableNames: 
    ' Debugging '
  poolDictionaries: ''
  category: 'Data Managers'
!


!ExternalDataManager class methodsFor: 'examples'!

example1
	"Fragments useful in testing."
	Junk := ExternalDataManager new.
	Junk openOn: 'junk.mng'.
	Junk close.
	Junk at: 1 put: 'Testing'.
	Junk at: 1.
	#('key1' 'key2' 'key3' 'key4' 'key5') do: [:key |
		| string |
		string := ''.
		100 timesRepeat: [string := string, key, ' '].
		Junk at: key put: string].
	ExternalDataManager debugging: true.
	ExternalDataManager debugging: false.
	#('key6' 'key2' 'key3' 'key7' 'key5') do: [:key |
		Junk at: key put: key].
	Junk keysAndValuesDo: [:key :value |
		MessageBox message: key printString, ' has value ', value printString].
	Junk at: 'key8' put: 'Try this with debugging true'.
	Junk at: 'key8' ifAbsent: ['Not there'].
	Junk compress.
	Smalltalk removeKey: #Junk
! !

!ExternalDataManager class methodsFor: 'querying'!
  
debugging: aBoolean
	Debugging := aBoolean
!
  
debugging
	^Debugging == true "Works even if not initialized"
!
   
headerSize
	^40 "bytes"
! !


!ExternalDataManager methodsFor: 'opening/closing'!
  
close
	"Writeout position of last dictionary entry and close the stream."
	stream isNil ifTrue: [^self]. "already closed"
	stream close.
	stream := nil.
!
 
openOn: fileName
	"Open a stream and read in data."
	| name |
	self alreadyOpen ifTrue: [^self].
	dictionary := Dictionary new.
	workStream := '' asStream.
	name := File fullPathName: fileName.
	(File exists: name)
		ifTrue: [
			stream := File pathName: name.
			self initializeOld]
		ifFalse: [
			stream := File newFile: name.
			self initializeNew].
! !

!ExternalDataManager methodsFor: 'accessing/modifying'!
  
at: key put: value
	self privateAt: key put: value.
	^value "maintain dictionary semantics"
!

at: key
	^self at: key ifAbsent: [self error: 'key ', key printString, ' does not exist']
!
   
at: key ifAbsent: aBlock
	"The value is in external memory. The private
	dictionary keeps track of its file position"
	| where |
	where := dictionary at: key ifAbsent: [^aBlock value].
	stream position: where.
	^ObjectFiler loadFrom: stream.
!
  
keysAndValuesDo: aBlock
	dictionary keys do: [:key |
		aBlock value: key value: (self at: key)]
! !

!ExternalDataManager methodsFor: 'querying'!
  
fileNameSuffix
	| name |
	name := self fileName.
	^name copyFrom: (File indexOfFilenameIn: name) to: name size
!

fileNamePrefix
	| name |
	name := self fileName.
	^name copyFrom: 1 to: (File indexOfFilenameIn: name) - 1
!

fileName
	self alreadyClosed ifTrue: [^''].
	^stream file pathName
! !

!ExternalDataManager methodsFor: 'compressing'!

compress
	"This operation is intended to be safe."
	| manager oldName newName |
	(MessageBox confirm: 'Begin Compress') ifFalse: [^self].
	oldName := self fileName.
	newName := self fileNamePrefix, '~$', self fileNameSuffix.
	manager := self class new.
	manager openOn: newName.
	self keysAndValuesDo: [:key :value |
		(manager privateAt: key put: value) ifFalse: [
			MessageBox message: 'Terminated unsuccessfully. Original unchanged'.
			manager close.
			File remove: newName.
			^self]].
	manager close. self close.
	File remove: oldName.
	File rename: newName to: oldName.
	self openOn: oldName.
	MessageBox message: 'Terminated successfully'
! !

!ExternalDataManager methodsFor: 'private initializing/finalizing'!
  
initializeNew
	self growBy: self class headerSize.
	last := 0.
	self finalize.
!

initializeOld
	"Previously constructed map."
	| key value workingLast valuePosition |
	stream isEmpty ifTrue: [
		self error: 'Not an external data manager file'].
	stream position: 0.
	last := workingLast := self load.
	[workingLast = 0] whileFalse: [
		stream position: workingLast.
		workingLast := self load. "previous one"
		key := self load.
		valuePosition := stream position. "but dont read"
		(dictionary includesKey: key) ifFalse: [
			dictionary at: key put: valuePosition]]
!
  
finalize
	stream position: 0.
	ObjectFiler dump: last on: stream.
	stream position: stream size; flush
! !

!ExternalDataManager methodsFor: 'private opening/closing support'!
   
alreadyClosed
	^stream isNil
!

alreadyOpen
	"Warns user and asks for directions. Returns true if already open."
	| reply name |
	self alreadyClosed ifTrue: [^false].
	name := stream file pathName.
	reply := MessageBox
		threeStateNotify: 'File "', name, '" already open!!'
		withText: 'Close and continue?'.
	reply == true ifTrue: [self close. ^false].
	reply == false ifTrue: [self error: 'file "', name, '" already open'].
	"Otherwise, it's nil."
	^true
! !

!ExternalDataManager methodsFor: 'private accessing/modifying'!
 
privateAt: key put: value
	"Returns whether or not entry was successfully inserted.
	Careful: dictionary contains key and stream position of value.
	Stream contains position of previous entry, key, and then value."
	| string1 string2 valuePosition |
	workStream reset; truncate.
	self dump: last; dump: key.
	string1 := workStream contents.
	workStream reset; truncate.
	self dump: value.
	string2 := workStream contents.
	self growBy: string1 size + string2 size
		ifSucceeded:  [:newLast |
			stream
				position: newLast;
				nextPutAll: string1.
			valuePosition := stream position.
			stream
				nextPutAll: string2.
			dictionary at: key put: valuePosition.
			last := newLast.
			self finalize.
			^true]
		ifFailed:  [^false "Don't append anything."]
! !

!ExternalDataManager methodsFor: 'private growing/loading/dumping'!
   
growBy: amount ifSucceeded: block1 ifFailed: block2
	"Go to the end of stream, output anything, and force to disk."
	| newLast |
	newLast := stream size.
	(self successfullyExecuted: [
		"Next line should be removed as soon as
		debugging is no longer needed."
		self class debugging ifTrue: [self smile].
		stream
			position: newLast + amount - 1;
			nextPut: $a; "anything at all"
			flush])
		ifTrue: [block1 value: newLast]
		ifFalse: block2
!
  
growBy: amount
	"Make sure user gets a message if there is no space."
	self growBy: amount ifSucceeded: [:newLast | ] ifFailed: []
!
 
load
	"Careful: Not the opposite of dump. Loads from
	the external stream, not the work stream."
	^ObjectFiler loadFrom: stream
!
   
dump: object
	"Careful: dumps only on the work stream."
	ObjectFiler dump: object on: workStream
! !

!ExternalDataManager methodsFor: 'private metalevel facilities'!
 
successfullyExecuted: aBlock
	"Only used when attempting to get NEW space."
	| reply |
	aBlock on: Error do: [:e |
		reply := MessageBox
			threeStateNotify: 'Out of space'
			withText: 'Delete unnecessary files and try again?'.
		reply == true ifTrue: [^self successfullyExecuted: aBlock].
		reply == false ifTrue: [self error: 'out of space'].
		"Otherwise, it's nil."
		^false].
	^true
! !
ExternalDataManager comment: ''!
